#include <windows.h>
#include <winsock2.h>
#include <stdlib.h>
#include <jni.h>
#include "de_NView_icmp_JPing.h"

#ifdef __cplusplus
extern "C" {
#endif

#define ICMP_ECHOREPLY                 0
#define ICMP_UNREACHABLE               3
#define ICMP_ECHO                      8
#define ICMP_CODE_NETWORK_UNREACHABLE  0
#define ICMP_CODE_HOST_UNREACHABLE     1

#define STATUS_FAILED       0xFFFF
#define PACKET_SIZE         32
#define ICMP_MIN            8

typedef struct
{
  unsigned int   h_len:4;
  unsigned int   version:4;
  unsigned char  tos;
  unsigned short total_len;
  unsigned short ident;
  unsigned short frag_and_flags;
  unsigned char  ttl;
  unsigned char  proto;
  unsigned short checksum;
  unsigned int   sourceIP;
  unsigned int   destIP;
}IP_HEADER;

typedef struct
{
  char           i_type;
  char           i_code;
  unsigned short i_cksum;
  unsigned short i_id;
  unsigned short i_seq;
  unsigned long  timestamp;
}ICMP_HEADER;

//Return:   delay (ms) on success,
//          -1L  UNREACHABLE
//          -2L  invalid/foreign Packet
long DecodePacket(char *buf,int bytes,unsigned short myId){
    IP_HEADER *ipHeader;
    ICMP_HEADER *icmpHeader;
    unsigned short ipHeaderLen;
    long rc;

    ipHeader=(IP_HEADER*)buf;
    ipHeaderLen=ipHeader->h_len*4;
    if(bytes<ipHeaderLen+ICMP_MIN){
        printf("Too few bytes\n");
        return -2L;
    }
    icmpHeader=(ICMP_HEADER*)(buf+ipHeaderLen);
    if(icmpHeader->i_id!=myId){return -2L;}     //not my packet
    if(icmpHeader->i_type!=ICMP_ECHOREPLY){
        if(icmpHeader->i_type==ICMP_UNREACHABLE){
            if(icmpHeader->i_code==ICMP_CODE_HOST_UNREACHABLE){
                return -1L;
            }
            if(icmpHeader->i_code==ICMP_CODE_NETWORK_UNREACHABLE){
                return -1L;
            }
        }
        else{   //invalid type
            return -2L;
        }
    }
    rc=GetTickCount()-icmpHeader->timestamp;
    if((rc>10000)||(rc<0))rc=-2L;   //invalid timestamp, maybe someone uses the same Id
    return rc;
}

unsigned short checksum(unsigned short *buffer,int size){
    unsigned long cksum=0;
    while(size>1){
        cksum+=*buffer++;
        size-=sizeof(unsigned short);
    }
    if(size){cksum+=*(unsigned char*)buffer;}
    cksum=(cksum>>16)+(cksum&0xffff);
    cksum+=(cksum>>16);
    return (unsigned short)(~cksum);
}

void SetupPacket(char * icmp_data,int datasize, unsigned short myId){
    ICMP_HEADER *icmp_hdr;
    char *datapart;

    icmp_hdr=(ICMP_HEADER*)icmp_data;
    icmp_hdr->i_type=ICMP_ECHO;
    icmp_hdr->i_code=0;
    icmp_hdr->i_id=myId;
    icmp_hdr->i_cksum=0;
    icmp_hdr->i_seq=0;
    datapart=icmp_data+sizeof(ICMP_HEADER);
    memset(datapart,'!',datasize-sizeof(ICMP_HEADER));
    memcpy(&datapart[3],"N-View",6);
}

long sendPacket(const char *str,unsigned short iD,SOCKET sockRaw){
    char icmp_data[PACKET_SIZE+sizeof(ICMP_HEADER)];
    HOSTENT* ptrHostent;
    SOCKADDR_IN addrDest;
    unsigned int addr=0;
    unsigned long timeout;

    ptrHostent=gethostbyname(str);
    if(!ptrHostent){addr=inet_addr(str);}
    if((!ptrHostent)&&(addr==INADDR_NONE)){
        printf("Error: Cannot resolve Host: %s\n",str);
        return -1L;
    }
    if(ptrHostent){
        memcpy(&(addrDest.sin_addr),ptrHostent->h_addr,ptrHostent->h_length);
        addrDest.sin_family=ptrHostent->h_addrtype;
    }
    else{
        addrDest.sin_addr.s_addr=addr;
        addrDest.sin_family=AF_INET;
    }
    SetupPacket(icmp_data,PACKET_SIZE+sizeof(ICMP_HEADER),(unsigned short) iD);
    ((ICMP_HEADER*)icmp_data)->i_cksum=0;
    ((ICMP_HEADER*)icmp_data)->timestamp=GetTickCount();
    ((ICMP_HEADER*)icmp_data)->i_seq=0;
    ((ICMP_HEADER*)icmp_data)->i_cksum=checksum((unsigned short*)icmp_data,PACKET_SIZE+sizeof(ICMP_HEADER));
    return sendto(sockRaw,icmp_data,PACKET_SIZE+sizeof(ICMP_HEADER),0,(SOCKADDR*)&addrDest,sizeof(addrDest));
}

#pragma argsused
JNIEXPORT jlong JNICALL 
Java_de_NView_icmp_JPing_sockStartup(JNIEnv *env,jobject obj){
    long rc;
    WORD wVersionRequested;
    WSADATA wsaData;
    SOCKET sockRaw;
    int timeoutMillis=1000;
    int bytesSent;
    wVersionRequested=MAKEWORD(2,1);
    rc=WSAStartup(wVersionRequested,&wsaData);
    if(rc==-1L)return rc;
    sockRaw=socket(AF_INET,SOCK_RAW,IPPROTO_ICMP);
    if(sockRaw==INVALID_SOCKET){
        printf("Error: Cannot create Socket\n");
        return -1L;
    }
    rc=setsockopt(sockRaw,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeoutMillis,sizeof(timeoutMillis));
    if(rc==SOCKET_ERROR){
        printf("Error: Cannot set Recv Timeout\n");
        closesocket(sockRaw);
        return -1L;
    }
    bytesSent=sendPacket("127.0.0.1",0,sockRaw);
    closesocket(sockRaw);
    if(bytesSent==SOCKET_ERROR){
        printf("Error: Cannot send packet\n");
        return -1L;
    }
    return 0L;
}

#pragma argsused
JNIEXPORT jlong JNICALL 
Java_de_NView_icmp_JPing_jPing(JNIEnv *env,jobject obj,jint timeoutMillis,jstring addrString,jint Id){
    const char* str;
    SOCKET sockRaw;
    SOCKADDR_IN addrFrom;
    int addrFromLen=sizeof(addrFrom);
    char recvbuf[PACKET_SIZE+sizeof(IP_HEADER)+sizeof(ICMP_HEADER)];
    int bytesReceived;
    int bytesSent;
    long rc;
    unsigned long timeout;

    if(addrString==NULL){
        printf("Error: NULL addrString!\n");
        closesocket(sockRaw);
        return -1L;
    }
    str=(*env)->GetStringUTFChars(env,addrString,NULL);
    if(str==NULL){
        printf("Error: Invalid addrString!\n");
        closesocket(sockRaw);
        return -1L;
    }
    sockRaw=socket(AF_INET,SOCK_RAW,IPPROTO_ICMP);
    if(sockRaw==INVALID_SOCKET){
        printf("Error: Cannot create Socket\n");
        return -1L;
    }
    rc=setsockopt(sockRaw,SOL_SOCKET,SO_RCVTIMEO,(char*)&timeoutMillis,sizeof(timeoutMillis));
    if(rc==SOCKET_ERROR){
        printf("Error: Cannot set Recv Timeout\n");
        closesocket(sockRaw);
        return -1L;
    }
    timeout=timeoutMillis+GetTickCount();
    bytesSent=sendPacket(str,Id,sockRaw);
    if(bytesSent==SOCKET_ERROR){
        printf("Error: Cannot send packet\n");
        closesocket(sockRaw);
        (*env)->ReleaseStringUTFChars(env, addrString,str );
        return -1L;
    }
    if(bytesSent!=(int)PACKET_SIZE+sizeof(ICMP_HEADER)){
        printf("Error: Too less bytes sent\n",str);
        closesocket(sockRaw);
        (*env)->ReleaseStringUTFChars(env, addrString,str );
        return -1L;
    }
    rc=-2L;
    while(rc==-2L){    //while receiving foreign packets
        bytesReceived=recvfrom(sockRaw,recvbuf,sizeof(recvbuf),0,(SOCKADDR*)&addrFrom,&addrFromLen);
        if(bytesReceived==SOCKET_ERROR){
            closesocket(sockRaw);
            (*env)->ReleaseStringUTFChars(env, addrString,str );
            return -1L;
        }
        if(timeout<GetTickCount()){
            closesocket(sockRaw);
            (*env)->ReleaseStringUTFChars(env, addrString,str );
            return -1L;
        }
        rc=DecodePacket(recvbuf,bytesReceived,(unsigned short) Id);
    }
    closesocket(sockRaw);
    (*env)->ReleaseStringUTFChars(env, addrString,str );
    return rc;
}
#ifdef __cplusplus
}
#endif
